unit oxoGrid;
{$I jb.inc}
interface

uses
  Classes, SysUtils, Forms, Grids, Windows, Controls, Messages,
  StdCtrls, Graphics, Clipbrd{$IFDEF VER17UP}, System.UITypes{$ENDIF};

type
  TOCopyStringGrid = class(TStringGrid)
  private
    procedure WMChar(var Msg: TWMChar); message WM_CHAR;
    procedure WMCopy(var Message: TMessage); message WM_COPY;
  end;

  TOScrollBarStringGrid = class(TOCopyStringGrid)
  private
    fPaintLock: Integer;
    fxScrollCounter: Integer;
    FScrollBars: {$IFDEF VER17UP}System.UITypes.{$ENDIF}TScrollStyle;
    
    procedure WMVScroll(var Msg: TWMScroll); message WM_VSCROLL;
    procedure WMHScroll(var Msg: TWMScroll); message WM_HSCROLL;

    procedure WMSize(var Message: TWMSize); message WM_SIZE;
    procedure CMShowingChanged(var Message: TMessage); message CM_SHOWINGCHANGED;

    procedure SetScrollBars(Value: {$IFDEF VER17UP}System.UITypes.{$ENDIF}TScrollStyle);
  protected
    procedure UpdateVerticalScrollBar; virtual;
    procedure UpdateHorizontalScrollBar; virtual;
    procedure UpdateScrollBars;

    procedure CreateParams(var Params: TCreateParams); override;
    procedure CreateWnd; override;

    procedure TopLeftChanged; override;
    procedure SizeChanged(OldColCount, OldRowCount: Longint); override;

    function DoMouseWheelDown(Shift: TShiftState; MousePos: TPoint): Boolean; override;
    function DoMouseWheelUp(Shift: TShiftState; MousePos: TPoint): Boolean; override;

    procedure ColWidthsChanged; override;
    procedure RowHeightsChanged; override;
  public
    constructor Create(AOwner: TComponent); override;
  public
    function LastVisibleCols: Integer;
    function LastVisibleRows: Integer;

    procedure Invalidate; override;
    procedure InvalidateCell(const aCol, aRow: Integer);
    procedure InvalidateCells(const aCellRect: TRect); overload;
    procedure InvalidateCells(const aLeftCol, aTopRow, aRightCol, aBottomRow: Integer); overload;

    procedure IncPaintLock;
    procedure DecPaintLock;
  published
    property TopRow;
    property LeftCol;
    
    property ScrollBars: TScrollStyle read FScrollBars write SetScrollBars default ssBoth;
  end;

implementation

uses Math;

{ TOCopyStringLineGrid }

procedure TOCopyStringGrid.WMChar(var Msg: TWMChar);
begin
  inherited;
  if Msg.CharCode = 3 then
    SendMessage(Handle, WM_COPY, 0, 0);
end;

procedure TOCopyStringGrid.WMCopy(var Message: TMessage);
begin
  Clipboard.AsText := (Cells[Col, Row]);
end;

procedure TOScrollBarStringGrid.CreateParams(var Params: TCreateParams); 
begin
  inherited CreateParams(Params);

  with Params do
  begin
    Style := Style or WS_TABSTOP;
    if FScrollBars in [ssVertical, ssBoth] then Style := Style or WS_VSCROLL;
    if FScrollBars in [ssHorizontal, ssBoth] then Style := Style or WS_HSCROLL;
  end;
end;

procedure TOScrollBarStringGrid.CreateWnd;
begin
  inherited;

  UpdateScrollBars;
end;

procedure TOScrollBarStringGrid.TopLeftChanged;
begin
  inherited;

  UpdateScrollBars;
end;

procedure TOScrollBarStringGrid.UpdateVerticalScrollBar;
var
  SIOld, SINew: TScrollInfo;
begin
  if Parent = nil then
    exit;

  SIOld.cbSize := sizeof(SIOld);
  SIOld.fMask := SIF_ALL;//SIF_ALL
  GetScrollInfo(Self.Handle, SB_VERT, SIOld);
  SINew := SIOld;

  SINew.nPage := LastVisibleRows;
  SINew.nMin := 0;
  SINew.nMax := RowCount-FixedRows-1;
  SINew.nPos := TopRow-FixedRows;

  if (SINew.nMin <> SIOld.nMin) or (SINew.nMax <> SIOld.nMax) or
   (SINew.nPage <> SIOld.nPage) or (SINew.nPos <> SIOld.nPos) then
    SetScrollInfo(Self.Handle, SB_VERT, SINew, True);
end;

procedure TOScrollBarStringGrid.UpdateHorizontalScrollBar;
var
  SIOld, SINew: TScrollInfo;
begin
  if Parent = nil then
    exit;

  SIOld.cbSize := sizeof(SIOld);
  SIOld.fMask := SIF_ALL;
  GetScrollInfo(Self.Handle, SB_HORZ, SIOld);
  SINew := SIOld;

  SINew.nPage := LastVisibleCols;
  SINew.nMin := 0;
  SINew.nMax := ColCount-FixedCols-1;
  SINew.nPos := LeftCol-FixedCols;

  if (SINew.nMin <> SIOld.nMin) or (SINew.nMax <> SIOld.nMax) or
   (SINew.nPage <> SIOld.nPage) or (SINew.nPos <> SIOld.nPos) then
    SetScrollInfo(Self.Handle, SB_HORZ, SINew, True);
end;

procedure TOScrollBarStringGrid.UpdateScrollBars;
begin
  UpdateVerticalScrollBar;
  UpdateHorizontalScrollBar;
end;

procedure TOScrollBarStringGrid.WMVScroll(var Msg: TWMScroll);
begin
  if Parent = nil then begin
    inherited;
    exit;
  end;

  case Msg.ScrollCode of
    SB_THUMBPOSITION,
    SB_THUMBTRACK: begin
      TopRow := Msg.Pos + FixedRows;
      Msg.Result := 0;
      fxScrollCounter := 0;
    end;
    SB_LINEUP: Dec(fxScrollCounter);
    SB_LINEDOWN: Inc(fxScrollCounter);
    SB_ENDSCROLL: begin
      //nothing
      TopRow := Max(FixedRows, Min(TopRow+fxScrollCounter, RowCount-LastVisibleRows));
      fxScrollCounter := 0;
    end;
  else
    inherited;
  end;
end;

procedure TOScrollBarStringGrid.WMHScroll(var Msg: TWMScroll);
begin
  if Parent = nil then begin
    inherited;
    exit;
  end;

  case Msg.ScrollCode of
    SB_THUMBPOSITION,
    SB_THUMBTRACK: begin
      LeftCol := Msg.Pos + FixedCols;
      Msg.Result := 0;
    end;
    //USE fxScrollCounter only for vertical scrolling!
    SB_LINELEFT: LeftCol := Max(FixedCols, Min(LeftCol-1, ColCount-LastVisibleCols));
    SB_LINERIGHT: LeftCol := Max(FixedCols, Min(LeftCol+1, ColCount-LastVisibleCols));
    SB_ENDSCROLL: //nothing
  else
    inherited;
  end;
end;

procedure TOScrollBarStringGrid.SizeChanged(OldColCount, OldRowCount: Longint);
begin
  inherited;

  UpdateScrollBars;
end;

procedure TOScrollBarStringGrid.WMSize(var Message: TWMSize);
begin
  inherited;
  
  UpdateScrollBars;
end;

procedure TOScrollBarStringGrid.CMShowingChanged(var Message: TMessage);
begin
  inherited;
  
  if Showing then begin
    UpdateScrollBars;
  end;
end;

function TOScrollBarStringGrid.DoMouseWheelDown(Shift: TShiftState;
  MousePos: TPoint): Boolean;
var NewTopRow: Integer;
begin
  NewTopRow := Max(FixedRows, Min(TopRow+3, RowCount-LastVisibleRows));
  result := NewTopRow <> TopRow;
  if result then
    TopRow := NewTopRow;

  UpdateScrollBars;
end;

function TOScrollBarStringGrid.DoMouseWheelUp(Shift: TShiftState;
  MousePos: TPoint): Boolean;
var NewTopRow: Integer;
begin
  NewTopRow := Max(FixedRows, Min(TopRow-3, RowCount-LastVisibleRows));
  result := NewTopRow <> TopRow;
  if result then
    TopRow := NewTopRow;

  UpdateScrollBars;
end;

procedure TOScrollBarStringGrid.Invalidate;
begin
  inherited;

  UpdateScrollBars;
end;

procedure TOScrollBarStringGrid.InvalidateCell(const aCol, aRow: Integer);
var R: TRect;
begin
  R := CellRect(aCol, aRow);
  Windows.InvalidateRect(Handle, @R, False);
end;

procedure TOScrollBarStringGrid.InvalidateCells(const aCellRect: TRect);
begin
  with aCellRect do
    InvalidateCells(Left, Top, Right, Bottom);
end;

procedure TOScrollBarStringGrid.InvalidateCells(const aLeftCol, aTopRow, aRightCol, aBottomRow: Integer);
var R: TRect;
begin
  R := BoxRect(aLeftCol, aTopRow, aRightCol, aBottomRow);
  Windows.InvalidateRect(Handle, @R, False);
end;

procedure TOScrollBarStringGrid.ColWidthsChanged;
begin
  inherited;

  UpdateScrollBars;
end;

procedure TOScrollBarStringGrid.RowHeightsChanged;
begin
  inherited;

  UpdateScrollBars;
end;

function TOScrollBarStringGrid.LastVisibleCols: Integer;
var I, Width: Integer;
begin
  result := 0;
  Width := ClientWidth + GridLineWidth;//posledni linka se ignoruje!
  for I := 0 to FixedRows-1 do//odecist FixedRows
    Dec(Width, ColWidths[I] + GridLineWidth);

  for I := ColCount-1 downto FixedCols do begin
    Dec(Width, ColWidths[I] + GridLineWidth);
    if Width < 0 then
      break;

    Inc(Result);

  end;
end;

function TOScrollBarStringGrid.LastVisibleRows: Integer;
var I, Height: Integer;
begin
  result := 0;
  Height := ClientHeight + GridLineWidth;//posledni linka se ignoruje!
  for I := 0 to FixedRows-1 do//odecist FixedRows
    Dec(Height, RowHeights[I] + GridLineWidth);

  for I := RowCount-1 downto FixedRows do begin
    Dec(Height, RowHeights[I] + GridLineWidth);
    if Height < 0 then
      break;

    Inc(Result);

  end;
end;

procedure TOScrollBarStringGrid.IncPaintLock;
begin
  Inc(fPaintLock);
  if HandleAllocated then
    SendMessage(Handle, WM_SetRedraw, 0, 0);
end;

procedure TOScrollBarStringGrid.DecPaintLock;
begin
  fPaintLock := Max(0, fPaintLock-1);
  if (fPaintLock = 0) and HandleAllocated then begin
    SendMessage(Handle, WM_SetRedraw, 1, 0);
    Invalidate;
  end;
end;

constructor TOScrollBarStringGrid.Create(AOwner: TComponent);
begin
  inherited;
  
  fPaintLock := 0;
  fxScrollCounter := 0;
  inherited ScrollBars := ssNone;
end;

procedure TOScrollBarStringGrid.SetScrollBars(Value: {$IFDEF VER17UP}System.UITypes.{$ENDIF}TScrollStyle);
begin
  FScrollBars := Value;
end;

end.